package br.usp.ime.algorithms;

import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;
import java.util.Vector;

import br.usp.ime.evolvinggraph.Edge;
import br.usp.ime.evolvinggraph.EvolvingGraph;
import br.usp.ime.evolvinggraph.Journey;
import br.usp.ime.evolvinggraph.Node;

public class OptimumJourneyAlgorithm implements Algorithm {
	
	Vector<Edge> eMin;
	Vector<Integer> tMin;
	double edgeCost, timeCost, journeyCost;
	public List<List <Journey> > pareto = new LinkedList < List < Journey> >();
	public List<List <Journey> > optimumJourney = new LinkedList < List < Journey> >();
	
	public List<List<Journey>> getOptimumJourney() {
		return optimumJourney;
	}

	public void setOptimumJourney(List<List<Journey>> optimumJourney) {
		this.optimumJourney = optimumJourney;
	}

	public double getJourneyCost() {
		return journeyCost;
	}

	public void setJourneyCost(double journeyCost) {
		this.journeyCost = journeyCost;
	}

	public double getTimeCost() {
		return timeCost;
	}

	public void setTimeCost(double timeCost) {
		this.timeCost = timeCost;
	}

	public double getEdgeCost() {
		return edgeCost;
	}

	public void setEdgeCost(double edgeCost) {
		this.edgeCost = edgeCost;
	}
	
	private boolean edgeAndScheduleSelection(EvolvingGraph graph, Vector<Integer> tLBD){
		Vector<Integer> arrivalTime = new Vector<Integer>();
		boolean improved = false;
		
		eMin = new Vector<Edge>(graph.getNumOfNodes());
		tMin = new Vector<Integer>(graph.getNumOfNodes());
		
		for (int i = 0; i < graph.getNumOfNodes(); i++){
			eMin.add(i, null);
			tMin.add(i, Integer.MAX_VALUE);
			arrivalTime.add(i, tLBD.get(i));
		}
		for (Node node : graph.getNodes()){
			if (tLBD.get(node.getId()) != Integer.MAX_VALUE){
				for (Edge edge : node.getAdj()){
					Integer newTime = edge.getNextDeparture(tLBD.get(node.getId()));
					if (newTime != null) {
						Node v = edge.getV();
						if ( newTime + edge.getCost() < arrivalTime.get(v.getId())){
							eMin.set(v.getId(), edge);
							tMin.set(v.getId(), newTime);
							arrivalTime.set(v.getId(), newTime + edge.getCost());
							improved = true;
						}
					}
				}
			}
		}
		return improved;
	}
	
	@Override
	public List<List<Journey>> calculate(EvolvingGraph graph, Node s) {
		return calculate(graph, s, 0);
	}
	
	public List<List<Journey>> calculate(EvolvingGraph graph, Node s, int initTime) {
		List<Journey> currentJourney = new ArrayList<Journey>();
		Vector<Integer> tLBD = new Vector<Integer>(graph.getNumOfNodes());
		
		for (int i = 0; i < graph.getNumOfNodes(); i++){
			tLBD.add(i, Integer.MAX_VALUE);
			currentJourney.add(i, new Journey());
			optimumJourney.add(i,new LinkedList<Journey>());
			pareto.add(i, new LinkedList<Journey>());
		}
		tLBD.set(s.getId(), initTime);
		currentJourney.set(s.getId(), new Journey());
		
		int numSteps = 0;
		
		while (numSteps < graph.getNumOfNodes()){
			List<Journey> lastJourney = new ArrayList<Journey>();
			if (edgeAndScheduleSelection(graph, tLBD) == false)
				break;
			
			for (int i = 0; i < graph.getNumOfNodes(); i++){
				lastJourney.addAll(currentJourney);
			}
			
			for (Node node : graph.getNodes()){
				setJourneyCost(Integer.MAX_VALUE);
				Edge edge = eMin.get(node.getId());
				if (edge != null){
					int uId = edge.getU().getId();
					int vId = edge.getV().getId();
					if(!optimumJourney.get(vId).isEmpty())
						setJourneyCost(optimumJourney.get(vId).get(0).getCost(timeCost, edgeCost));
					int t = tMin.get(vId);
					Journey journey = new Journey(lastJourney.get(uId), edge, t);
					pareto.get(vId).add(journey);
					double costCurrentJourney;
					costCurrentJourney = getEdgeCost()*journey.getSize() + getTimeCost()*journey.getArrivalDate();
					if(getJourneyCost() > costCurrentJourney){
						optimumJourney.get(vId).clear();
						optimumJourney.get(vId).add(journey);
						setJourneyCost(costCurrentJourney);
					}
					else if(getJourneyCost() == costCurrentJourney){
						optimumJourney.get(vId).add(journey);
					}
					journey.getArrivalDate();
					journey.getSize();
					
					currentJourney.set(vId, journey);
					tLBD.set(vId, t + edge.getCost());
				}
			}
			numSteps = numSteps + 1;
		}
		
		return optimumJourney;
	}

	public List<List<Journey>> getPareto() {
		return pareto;
	}

	public void setPareto(List<List<Journey>> pareto) {
		this.pareto = pareto;
	}
	
	public static void main(String args[]) throws FileNotFoundException{
		EvolvingGraph eg = EvolvingGraph.readEvolvingGraph("input/teste7.in");
		OptimumJourneyAlgorithm algorithm = new OptimumJourneyAlgorithm();
		algorithm.setEdgeCost(6);
		algorithm.setTimeCost(1);
		algorithm.setJourneyCost(Integer.MAX_VALUE);
		algorithm.calculate(eg,eg.getNode(0));
		List<List<Journey>> pareto = algorithm.getPareto();
		List<List<Journey>> optimumJourney = algorithm.getOptimumJourney();
		int j = 11;
		System.out.println("Pontos da curva de pareto :");
		for(int i = 0; i < pareto.get(j).size(); i++){
			System.out.println(pareto.get(j).get(i).getSize() + " " + pareto.get(j).get(i).getArrivalDate());
		}
		System.out.println("Jornada de menor custo: ");
		System.out.println(optimumJourney.get(j));
		System.out.println("Custo da jornada ótima escolhida: " + optimumJourney.get(j).get(0).getCost(algorithm.getTimeCost(), algorithm.getEdgeCost()));
		
	}

}
