package br.usp.ime.evolvinggraph;


import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Scanner;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.regex.MatchResult;

public class EvolvingGraph {
	List<Node> nodes;
	int edgesnum;
	
	public EvolvingGraph(int numOfNodes){
		nodes = new ArrayList<Node>();
		edgesnum = 0;
		for(int i = 0; i < numOfNodes; i++){
			nodes.add(new Node(i));
		}
	}

	public List<Node> getNodes() {
		return nodes;
	}
	
	public Node getNode(int id){
		return nodes.get(id);
	}
	
	public int getNumOfNodes() {
		return nodes.size();
	}
	
	public void createEdge(int u, int v, SortedSet<Integer> sched, int cost) {
		nodes.get(u).add(this.nodes.get(v), sched, cost);
		edgesnum++;
	}
	
	public Edge getEdge (int u, int v) {
		Edge e = null;
		for (Iterator<Edge> iter = this.getNode(u).getAdj().iterator(); iter.hasNext();) {
			Edge element = (Edge) iter.next();
			if (element.getV().getId() == v) {
				e = element;
				break;
			}
		}
		return e;
	}
	
	public int getNumOfEdges () {
		return this.edgesnum;
	}
	
	public String toString(){
		String str = this.getNumOfNodes() + " " + edgesnum + "\n" ;
		for (int i = 0; i < nodes.size(); i++){
			Node u = nodes.get(i);
			for (int j = 0; j < u.adj.size(); j++){
				Edge e = u.adj.get(j);
				str += u.id + " " + e.v.id + " " + e.cost + " " + e.schedule.size();
				Iterator<Integer> it = e.schedule.iterator();
				str += " " + it.next();
				while(it.hasNext()){
					str += " " + it.next();
				}
				str += "\n";
			}
		}
		return str;
	}
	
	public static EvolvingGraph readEvolvingGraph (String filename) {
		SortedSet<Integer> sched;
		int source, destiny, cost, numsched;
		int aux;
		File f = new File (filename);
		EvolvingGraph grafo = new EvolvingGraph(0);
		
		try {
			Scanner sc = new Scanner (f);
			int nodes = sc.nextInt();
			int edges = sc.nextInt();
			grafo = new EvolvingGraph(nodes);
			while(edges > 0) {
				source = sc.nextInt();
				destiny = sc.nextInt();
				cost = sc.nextInt();
				numsched = sc.nextInt();
				sched = new TreeSet<Integer>();
				while(numsched > 0) {
					aux = sc.nextInt();
					sched.add(aux);
					numsched--;
				}
				grafo.createEdge(source, destiny, sched, cost);
				edges--;
			}
		}
		catch (Exception e){
			System.out.println(e);
			System.out.println ("Erro na leitura do grafo. F**eu!");
		}
		return grafo;
	}
	
    private static void processLine(Scanner s, EvolvingGraph eg, HashSet<Edge> openConnections) {
    	SortedSet<Integer> sched, schedc;
        s.findInLine("(\\d+).00 CONN (\\d+) (\\d+) (\\w+)");
        MatchResult result = s.match();
        int time = Integer.parseInt((result.group(1)));
        int u = Integer.parseInt((result.group(2)));
        int v = Integer.parseInt((result.group(3)));
        String status = result.group(4);
        Edge uv = null, vu = null;
        uv = eg.getEdge(u, v);
        if (uv == null) {
        	sched = new TreeSet<Integer> ();
        	sched.add(time);
        	schedc = new TreeSet<Integer> ();
        	schedc.add(time);
        	eg.createEdge(u, v, sched, 0);
        	eg.createEdge(v, u, schedc, 0);
        	openConnections.add(eg.getEdge(u, v));
        	openConnections.add(eg.getEdge(u, v));
        }
        else {
        	vu = eg.getEdge(v, u);
        	sched = uv.getSchedule();
    		schedc = vu.getSchedule();
        	if (status.equals("up")) {
        		sched.add(time);
        		schedc.add(time);
        		openConnections.add(uv);
        		openConnections.add(vu);
        	}
        	else {
        		int ini = sched.last();
        		openConnections.remove(uv);
        		openConnections.remove(vu);
        		for (int i = ini + 1; i < time; i++) {
        			sched.add(i);
        			schedc.add(i);
        		}
        	}
        }
        
        //System.out.println(time + " " + u + " " + v + " " + status);
    }
    
	public static EvolvingGraph readEvolvingGraphFromONE (String filename, int numOfNodes, int simTime) {
		EvolvingGraph graph = new EvolvingGraph(numOfNodes);
		HashSet<Edge> openConnections = new HashSet<Edge> ();
		File f = new File (filename);
		Scanner sc;
		
		try {
			sc = new Scanner (f);
			while (sc.hasNextLine ()) {
				processLine(sc, graph, openConnections);
				sc.nextLine();
			}
			for (Iterator <Edge>iter = openConnections.iterator(); iter.hasNext();) {
				Edge element = (Edge) iter.next();
				int ini = element.getSchedule().last();
				for (int i = ini + 1; i < simTime; i++)
					element.getSchedule().add(i);
			}
		}
		catch (FileNotFoundException e) {
			System.out.println ("Erro na leitura do grafo.");
		}
		return graph;
	}

	public void removeJourney(Journey j) {
		ArrayList<Edge> edges = j.getEdges();
		ArrayList<Integer> times = j.getTimes();
		for (int i = 0; i < edges.size(); i++) {
			Edge e = edges.get(i);
			int t = times.get(i);
			SortedSet<Integer> s = e.getSchedule();
			s.remove(t);
			if (s.isEmpty()) {
				Node u = e.getU(), v = e.getV();
				u.getAdj().remove(e);
				v.getAdj().remove(e);
			}
			
		}
	}
}
